function [W,BETA,gamma,alpha,perr,err] = proDSfit(X,S,W0,BETA0,gamma0,alpha0,options,Optimp,Visu);

% Copyright Thierry Denoeux 
% February 14, 2001

% Learning procedure for Belief Structure Neural Network
%
% [W,BETA,gamma,alpha,perr,err] = proDSfit(X,S,W0,BETA0,gamma0,alpha0,options,Optimp,Visu);
%
% Input :
%
% X    : data matrix (nb samples x p)
% S    : class labels  (nb samples x 1) ; maximum = M = number of classes  
% W0 : n prototypes, size (n x p)   
% BETA0 : parameter used to compute membership degrees, size (n,M)  
% gamma0 : defines influence regions; size = (n,1)  
% alpha0 : size (n,1)
% options : parameter vector (may be empty)
%
% options(1) = 1 if normalization, 0 otherwise (default=1)
% options(2) = lambda (default=1/nb classes)
% options(3) = regularization parameter (default=0)
% options(4) = 1 if optimization of prototype centers, 0 otherwise (default=1)
%
% Optimp : optimization parameter (may be empty)  
% Optimp(1) : maximum number of iterations (default=500)
% Optimp(2) : stopping threshold (default = 1e-4)
%
% Visu = visualization parameter ( may be empty, default = 0)
% Visu = 0 : no visualization
% Visu = k : visualization every k iterations
%
% All arguments must be supplied !
%
% Output :
%
% optimized parameters, and
% perr = training set estimate of error probability
% err = empirical error criterion
%
% Reference:
%
% T. Denoeux. A neural network classifier based on Dempster-Shafer theory.
% IEEE Trans. on Systems, Man and Cybernetics A, 30(2):131-150, 2000.

M=max(S);

if nargin < 9,
   disp('ERROR: not enough input arguments !')
   return;
end;

if isempty(options),
  options=[1 1/M 0 1];
end;

if isempty(Optimp),
  Optimp=[500 1e-4];
end;

if isempty(Visu),
  Visu=0;
end;


[n,p]=size(W0);

Id = eye(M);
t=Id(S,:);
N=size(X,1);

%--------------------------------------------------------------
% Optimisation
%--------------------------------------------------------------
w = [reshape(W0,n*p,1)
     gamma0
     reshape(BETA0,n*M,1)
     alpha0];

[w,E,err] = harris(w,X',t',options,Optimp,Visu);

l=length(w);
W=reshape(w(1:n*p),n,p);
gamma = w(n*p+1 : n*p+n);
BETA = reshape(w((n*(p+1)+1):n*(p+1+M)),n,M);
alpha = reshape(w(n*(p+1+M)+1:l),n,1);
BETA2 = BETA .^2;
beta2 = sum(BETA2')';
U = BETA2 ./ (beta2*ones(1,M));

%-----------------------------------------------------------------
% calcul du taux d'erreur
%-----------------------------------------------------------------
mk = [zeros(M,N);ones(1,N)];
for k=1:n,
  d = 0.5*sum((X' - W(k,:)'*ones(1,N)).^2);
  alphap(k)=0.99 ./ (1 + exp(-alpha(k)));
  s = alphap(k)*exp(- gamma(k).^2 * d);
  m = [U(k,:)'*s ; 1-s];
  mk = [ (mk(1:M,:).*(m(1:M,:) + ones(M,1)*m(M+1,:)) + ...
m(1:M,:).*(ones(M,1)*mk(M+1,:)) ) ; (mk(M+1,:).*m(M+1,:))]; 
  end;

[mmax,imax] = max(mk(1:M,:));

perr = length(find(imax ~= S'))/N;


%%%%%%%%%%%%%%%%%%%%%%% Routine Harris %%%%%%%%%%%%%%%%%%%%%%%
function [xp,yp,err] = harris(x0,z,t,options,Optimp,Visu)
%
% optimisation par la methode de Harris
%
% fonc: la fonction a optimiser (renvoie aussi le gradient)
% x0: etat initial
% z: les autres parametres de fonc et grad (entrees reseau)
% t: les autres parametres de fonc et grad (sorties reseau)
%

inorm=options(1);
lambda=options(2);
mu=options(3);
iw = options(4);

pas = 0.1 * ones(size(x0));
a = 1.2;
b = 0.8;
c = 0.5;
ovf = 1e4 ;
unf = 1e-6;
x=x0;

err=[];
it = 0;
gain=1;
if inorm,
  [yp,gp]=foncgradr2n(x,z,t,options);
else
  [yp,gp]=foncgradr2(x,z,t,options);
end;

while (gain >= Optimp(2)) & (it <= Optimp(1))
  it=it+1;
  if inorm,
    [y,g]=foncgradr2n(x,z,t,options);
  else
    [y,g]=foncgradr2(x,z,t,options);
  end;
  err(it)=y;
  if Visu >0,
    if rem(it,Visu)==1, 
      disp([it y 10*gain]); 
    end;
  end;
  if y > yp,    
    x = xp;
    g = gp; 
    pas = pas * c;
    x = x - pas .* g;
  else
    gain=.9*gain+.1*abs(yp-y);
    xp = x;
    test = (g .* gp) >= 0;
    pas = ((test * a) + ((~test) * b)) .* pas;
    pas = (pas<=ovf) .* pas + (pas>ovf) * ovf;
    pas = (pas>=unf) .* pas + (pas<unf) * unf;
    gp = g;
    x = x - pas .* g;
    yp = y;
  end;
end




%%%%%%%%%%%%%%%%%%%%%%%%%%%% Routine foncgradr2 %%%%%%%%%%%%%%%%%

function [E,G] = foncgradr2(w,X,t,options);

lambda=options(2);
mu=options(3);
iw = options(4);

l = length(w);
[p,N]=size(X);
[M,N]=size(t);
n = l / (p+2+M);

W=reshape(w(1:n*p),n,p);
gamma = w(n*p+1 : n*p+n);
BETA = reshape(w((n*(p+1)+1):n*(p+1+M)),n,M);
alpha = reshape(w(n*(p+1+M)+1:l),n,1);
BETA2 = BETA .^2;
beta2 = sum(BETA2')';
U = BETA2 ./ (beta2*ones(1,M));

d=zeros(n,N);
s=d;
expo=s;
dU = zeros(n,M);
Dbeta = dU;
Ds = zeros(n,N);
Dgamma = Ds;
Dalpha = Ds;
mk = [zeros(M,N);ones(1,N)];
mm = mk;
alphap = 0.99 ./ (1 + exp(-alpha));


for k=1:n,
% propagation
  d(k,:) = 0.5*sum((X - W(k,:)'*ones(1,N)).^2);
  expo(k,:) = exp(- gamma(k).^2 * d(k,:));
  s(k,:) = alphap(k)*expo(k,:);
  m = [U(k,:)'*s(k,:) ; 1-s(k,:)];
  mk = [ (mk(1:M,:).*(m(1:M,:) + ones(M,1)*m(M+1,:)) + ...
         m(1:M,:).*(ones(M,1)*mk(M+1,:)) ) ; (mk(M+1,:).*m(M+1,:))];
  end;
Q = mk(1:M,:) + (lambda * ones(M,1)*mk(M+1,:)) - t;
E = 0.5 * mean(sum(Q.^2)) + mu * sum(alphap);

for k=1:n,
% calcul gradient
  m = [U(k,:)'*s(k,:) ; 1-s(k,:)];
  mm(M+1,:) = mk(M+1,:) ./ m(M+1,:);
  L = (ones(M,1)*mm(M+1,:));
  mm(1:M,:) = (mk(1:M,:) - L .* m(1:M,:)) ./ (m(1:M,:) + ones(M,1)*m(M+1,:));
  R = mm(1:M,:) + L;
  A = R .* (ones(M,1)*s(k,:));
  B = (U(k,:)'*ones(1,N)) .* R - mm(1:M,:) - lambda * L;
  dU(k,:) = mean((A .* Q)');
  Ds(k,:) = sum(Q .* B);
  DW(k,:) = - (Ds(k,:) .* (gamma(k).^2 * ones(1,N)) .* s(k,:)) * ...
                                    (ones(N,1)*W(k,:) - X');
  end;
DW = iw*DW/N;
T = beta2*ones(1,M);
Dbeta = 2*BETA ./ T.^2 .* (dU .*(T - BETA2) - ...
                      (sum((dU.*BETA2)'))'*ones(1,M) + dU .* BETA2);

Dgamma = - 2 * (mean((Ds .* d .* s)'))' .* gamma;
Dalpha = ((mean((Ds .* expo)'))'+ mu) .* (0.99 * (1-alphap) .* alphap);

G = [reshape(DW,n*p,1)
     Dgamma
     reshape(Dbeta,n*M,1)
     Dalpha];



%%%%%%%%%%%%%%%%%%%%%%%%%%%% Routine foncgradr2n %%%%%%%%%%%%%%%%%

function [E,G] = foncgradr2n(w,X,t,options);
% version avec normalisation

lambda=options(2);
mu=options(3);
iw = options(4);

l = length(w);
[p,N]=size(X);
[M,N]=size(t);
n = l / (p+2+M);

W=reshape(w(1:n*p),n,p);
gamma = w(n*p+1 : n*p+n);
BETA = reshape(w((n*(p+1)+1):n*(p+1+M)),n,M);
alpha = reshape(w(n*(p+1+M)+1:l),n,1);
BETA2 = BETA .^2;
beta2 = sum(BETA2')';
U = BETA2 ./ (beta2*ones(1,M));

d=zeros(n,N);
s=d;
expo=s;
dU = zeros(n,M);
Dbeta = dU;
Ds = zeros(n,N);
Dgamma = Ds;
Dalpha = Ds;
DW=zeros(size(W));
mk = [zeros(M,N);ones(1,N)];
mm = mk;
alphap = 0.99 ./ (1 + exp(-alpha));

for k=1:n,
% propagation
  d(k,:) = 0.5*sum((X - W(k,:)'*ones(1,N)).^2);
  expo(k,:) = exp(- gamma(k).^2 * d(k,:));
  s(k,:) = alphap(k)*expo(k,:);
  m = [U(k,:)'*s(k,:) ; 1-s(k,:)];
  mk = [ (mk(1:M,:).*(m(1:M,:) + ones(M,1)*m(M+1,:)) + ...
         m(1:M,:).*(ones(M,1)*mk(M+1,:)) ) ; (mk(M+1,:).*m(M+1,:))];
  end;
% normalisation
K = sum(mk);K2=K.^2;
mkn = mk ./ (ones(M+1,1)*K); 
Q = mkn(1:M,:) + (lambda * ones(M,1)*mkn(M+1,:)) - t;
E = 0.5 * mean(sum(Q.^2)) + mu * sum(alphap);

dEdm = zeros(M+1,N);
I = eye(M);
for k=1:M,
 dEdm(k,:) = sum( Q .* (I(:,k)*K - mk(1:M,:) - lambda*ones(M,1)*mk(M+1,:))) ./ K2;
 end;
dEdm(M+1,:) = sum(Q .* (-mk(1:M,:) + lambda*ones(M,1)*(K-mk(M+1,:)))) ./ K2;

for k=1:n,
% calcul gradient
  m = [U(k,:)'*s(k,:) ; 1-s(k,:)];
  mm(M+1,:) = mk(M+1,:) ./ m(M+1,:);
  L = (ones(M,1)*mm(M+1,:));
  mm(1:M,:) = (mk(1:M,:) - L .* m(1:M,:)) ./ (m(1:M,:) + ones(M,1)*m(M+1,:));
  R = mm(1:M,:) + L;
  A = R .* (ones(M,1)*s(k,:));
  B = (U(k,:)'*ones(1,N)) .* R - mm(1:M,:);
  dU(k,:) = mean((A .* dEdm(1:M,:))');
  Ds(k,:) = sum(dEdm(1:M,:) .* B) - (dEdm(M+1,:).*mm(M+1,:));
  DW(k,:) = - (Ds(k,:) .* (gamma(k).^2 * ones(1,N)) .* s(k,:)) * (ones(N,1)*W(k,:) - X');
  end;
DW = iw*DW/N;

T = beta2*ones(1,M);
Dbeta = 2*BETA ./ T.^2 .* (dU .*(T - BETA2) - ...
                      (sum((dU.*BETA2)'))'*ones(1,M) + dU .* BETA2);

Dgamma = - 2 * (mean((Ds .* d .* s)'))' .* gamma;
Dalpha = ((mean((Ds .* expo)'))'+ mu) .* (0.99 * (1-alphap) .* alphap);


G = [reshape(DW,n*p,1)
     Dgamma
     reshape(Dbeta,n*M,1)
     Dalpha];

